#pragma rtGlobals=1		// Use modern global access method.

// ----------------- Neuralynx Viewer -----------------------

// Initialize the Neuralynx Viewer.  
Function InitNlxViewer()
	String curr_folder=GetDataFolder(1)
	NewDataFolder /O/S root:Packages
	NewDataFolder /O/S Nlx
	Variable /G spikeNum=0, YAxisRange=0
	SetDataFolder $curr_folder
End

// Views Neuralynx spike data.  
Function CreateNlxViewer(Data,type)
	Wave Data
	String type // "ntt" or "ncs" (the latter is not yet supported for viewing).  
	
	String folder="root:Packages:Nlx:"
	InitNlxViewer()
	DoWindow /K NlxViewer
	Display /N=NlxViewer /K=1 /W=(100,100,600,600)
	ControlBar /L 130
	Button ControlToggle, pos={0,0}, size={25,20}, proc=NlxViewerButtons, title="-", userData="1"
	Variable y_pos=25, y_space=25
	String name=NameOfWave(data)
	String available=NlxAvailableWaves(type)
	PopupMenu DataName, value=#("NlxAvailableWaves(\""+type+"\")"), mode=(1+WhichListItem(name,available)), pos={2,y_pos}, proc=NlxViewerPopupMenus
	y_pos+=y_space
	SetVariable SpikeNum, title="Spike #", value=$(folder+"spikeNum"), bodywidth=50, pos={2,y_pos}, proc=NlxViewerSetVariables
	Checkbox AllSpikes, title="All", pos={95,y_pos+2}, proc=NlxViewerCheckboxes
	y_pos+=y_space
	PopupMenu Cluster, title="Cluster", pos={2,y_pos+2}, value=#"NlxClusterChoices()",proc=NlxViewerPopupMenus
	y_pos+=y_space
	SetVariable YAxisRange, title="Range (\F'Symbol'm\F'Arial'V)", value=$(folder+"YAxisRange"), limits={0,Inf,10}, bodywidth=50, pos={2,y_pos}, proc=NlxViewerSetVariables
	y_pos+=y_space
	String fullPath=GetWavesDataFolder(Data,2)
	ValDisplay TimeStamp, title="Timestamp", bodywidth=50, pos={2,y_pos}, value=#(fullPath+"_t[0]/1000000")
	Variable i
	ReplaceNlxWaves(Data,type)
	strswitch(type)
		case "ntt":
			break
		case "ncs": // ncs not supported for viewing using this function.  
			break
	endswitch
End

// Used to build the popup menu for Neuralynx clusters.  
Function /S NlxClusterChoices()
	ControlInfo DataName
	Wave Data=$S_Value
	Wave Clusters=$(GetWavesDataFolder(Data,2)+"_c")
	Variable maxx=wavemax(Clusters)
	Variable i
	String choices="All;Assigned;"
	for(i=0;i<=maxx;i+=1)
		choices+=num2str(i)+";"
	endfor
	return choices
End

//  Uses to build the popu menu for Neuralynx data waves.  
Function /S NlxAvailableWaves(type)
	String type
	String list=""
	Variable i
	strswitch(type)
		case "ntt":
			list=WaveList(type+"*",";","DIMS:3,MINLAYERS:4,MAXLAYERS:4")
			break
		default:
			for(i=0;i<100;i+=1)
				String candidate=type+num2str(i)
				if(exists(candidate))
					list+=candidate+";"
				endif
			endfor
		break
	endswitch
	return list
End

// Replaces the data in the Neuralynx viewer with a new data wave (usually a different electrode).  
Function ReplaceNlxWaves(Data,type)
	Wave Data // The new data wave.  
	String type // Type of the new data wave.  
	
	String fullPath=GetWavesDataFolder(Data,2)
	SetWindow kwTopWin userData(fullPath)=fullPath
	Wave /Z Clusters=$(fullPath+"_c")
	ControlInfo Cluster; String cluster=S_Value
	Variable red,green,blue
	NlxCluster2Color(cluster,wavemax(Clusters),red,green,blue)
	strswitch(cluster)
		case "All":
			Extract /O/INDX Clusters, TempClusters, 1 // All spikes.  
			break
		case "Assigned":
			Extract /O/INDX Clusters, TempClusters, Clusters!=0 && numtype(Clusters)==0
			break
		default:
			Extract /O/INDX Clusters, TempClusters, Clusters==str2num(cluster)
			break
	endswitch
	
	// Remove old traces.  
	String traces=TraceNameList("",";",3)
	traces=SortList(traces,";",17)
	Variable i,j
	for(i=0;i<ItemsInList(traces);i+=1)
		RemoveFromGraph $StringFromList(i,traces)
	endfor
		
	ControlInfo /W=NlxViewer AllSpikes; Variable allSpikes=V_Value
	ControlInfo yAxisRange; Variable yAxisRange=V_Value
	
	strswitch(type)
		case "ntt":
			for(i=0;i<4;i+=1)
				String axis_name="axis_"+num2str(i)
				String axisT_name="axisT_"+num2str(i)
				for(j=0;j<(allSpikes ? numpnts(TempClusters) : 1);j+=1) // Append all spikes on this channel or only the first one.  
					Variable index=TempClusters[j]
					AppendToGraph /c=(red,green,blue) /L=$axis_name /B=$axisT_name Data[][index][i]
				endfor
				ModifyGraph /Z axisEnab($axis_name)={(i<2)*0.52,0.48+(i<2)*0.5}, freePos($axis_name)={0,$axisT_name}
				ModifyGraph /Z axisEnab($axisT_name)={(mod(i,2)==1)*0.52,0.48+(mod(i,2)==1)*0.5}, freePos($axisT_name)={0,$axis_name}
				ModifyGraph /Z tickUnit($axis_name)=1,prescaleExp($axis_name)=6,btlen=1,fsize=9
				SetAxis /Z $axis_name -yAxisRange/1000000,yAxisRange/1000000
				Label /Z $axis_name " "
			endfor
			ValDisplay Timestamp value=#(fullPath+"_t[0]/1000000")
			break
		case "nse":
			AppendToGraph Data[][0]
			ModifyGraph tickUnit(left)=1,prescaleExp(left)=6
			Label left "\F'Symbol'm\F'Arial'V"
			break
	endswitch
	KillWaves /Z TempClusters
End

// Returns a color by reference for a given spike cluster number.  
Function NlxCluster2Color(cluster,numClusters,red,green,blue)
	String cluster
	Variable numClusters,&red,&green,&blue
	
	strswitch(cluster)
		case "All":
			red=0; green=0; blue=0
			break
		case "Assigned":
			red=10000; green=10000; blue=10000
			break
		default:
			ColorTab2Wave rainbow; Wave M_Colors
			Variable clusterChoice=str2num(cluster)
			Variable colorIndex=clusterChoice*dimsize(M_Colors,0)/numClusters
			red=M_Colors[colorIndex][0]; green=M_Colors[colorIndex][1]; blue=M_Colors[colorIndex][2]
			break
	endswitch
	KillWaves /Z M_Colors
End

Function NlxViewerButtons(info)
	Struct WMButtonAction &info
	if(info.eventCode!=2)
		return 0
	endif
	Variable on=str2num(GetUserData("","ControlToggle",""))
	if(on)
		ModifyControlList ControlNameList("", ";", "*") disable=1
		Button ControlToggle title="+", userData="0", disable=0
		ControlBar /L 0
		ModifyControlList ControlNameList("", ";", "*_tab0") disable=1
	else
		ModifyControlList ControlNameList("", ";", "*") disable=0
		Button ControlToggle title="-", userData="1"
		ControlBar /L 120
	endif
End

Function NlxViewerCheckboxes(ctrlName,checked)
	String ctrlName
	Variable checked
	
	ControlInfo DataName
	String dataName=S_Value
	Wave Data=$dataName
	String type=NlxType(Data)
	if(checked)
		SetVariable SpikeNum, disable=2
		ValDisplay TimeStamp, disable=1
		ReplaceNlxWaves(Data,type)
	else
		SetVariable SpikeNum, disable=0
		ValDisplay TimeStamp, disable=0
		ReplaceNlxWaves(Data,type)
	endif
End

Function NlxViewerSetVariables(info)
	Struct WMSetVariableAction &info
	
	if(info.eventCode<0)
		return 0
	endif
	ControlInfo DataName
	String dataName=S_Value
	Wave Data=$dataName
	
	strswitch(info.ctrlName)
		case "SpikeNum":
			Variable spikeNum=info.dval
			ControlInfo Cluster; String cluster=S_Value
			Wave Clusters=$(GetWavesDataFolder(Data,2)+"_c")
			Variable red,green,blue
			NlxCluster2Color(cluster,wavemax(Clusters),red,green,blue)
			strswitch(cluster)
				case "All":
					// Do nothing.  
					break
				case "Assigned":
					Extract /O/INDX Clusters, TempClusters, Clusters!=0 && numtype(Clusters)==0
					spikeNum=TempClusters[spikeNum]
					break
				default:
					Extract /O/INDX Clusters, TempClusters, Clusters==str2num(cluster)
					spikeNum=TempClusters[spikeNum]
					break
			endswitch
			Variable i
			String traces=TraceNameList("",";",3)
			for(i=0;i<ItemsInList(traces);i+=1)
				String trace_name=StringFromList(i,traces)
				Wave Data=TraceNameToWaveRef("", trace_name)
				ReplaceWave trace=$trace_name Data[][spikeNum][i] // Don't need to use ReplaceNlxWaves because this is the same data as before.  
			endfor
			ModifyGraph /Z rgb=(red,green,blue)
			String data_path=GetWavesDataFolder(Data,2)
			ValDisplay TimeStamp, value=#(data_path+"_t["+num2str(spikeNum)+"]/1000000")
			
			KillWaves /Z TempClusters,M_Colors
			break
		case "YAxisRange":
			Variable y_range=info.dval
			String axes=AxisList("")
			for(i=0;i<ItemsInList(axes);i+=1)
				String axis_name=StringFromList(i,axes)
				String axis_info=AxisInfo("",axis_name)
				String type=StringByKey("AXTYPE",axis_info)
				if(StringMatch(type,"LEFT"))
					if(y_range==0)
						SetAxis /A $axis_name
					else
						SetAxis $axis_name -y_range/1000000,y_range/1000000
					endif
					String axisT_name=ReplaceString("axis",axis_name,"axisT")
					SetAxis /A $axisT_name
				endif
			endfor
			break
	endswitch
End

Function NlxViewerPopupMenus(info)
	Struct WMPopupAction &info
	if(info.eventCode!=2)
		return 0
	endif
	Variable i
	ControlInfo DataName
	Wave Data=$S_Value
	String type=NlxType(Data) // Works for now, but will need to be changed.  
	strswitch(info.ctrlName)
		case "DataName":
			ReplaceNlxWaves(Data,type)
			PopupMenu Cluster, mode=1
			break
		case "Cluster":
			ControlInfo AllSpikes
			if(V_Value)
				ReplaceNlxWaves(Data,type)
			else
				Struct WMSetVariableAction info2; info2.eventCode=1; info2.ctrlName="SpikeNum"
				NlxViewerSetVariables(info2)
			endif
			break
	endswitch
End

// --------------------------------------------------------------------------------------------------------------

// Returns the neuralynx data type "ntt", "ncs", etc. for a given data wave.  Only "ntt" supported and only by virtue of its 4 layer (tetrode) nature.  
Function /S NlxType(Data)
	Wave Data
	if(dimsize(Data,2)==4)
		return "ntt"
	else
		return ""
	endif
End

Function NlxPCA(Data[,flatten])
	Wave Data // e.g. a 32xNx4 Tetrode wave.  
	Variable flatten // Flatten a tetrode into a single waveform with 4 times the length.  
	
	Variable electrode, num_electrodes=max(1,dimsize(Data,2))
	Variable points=dimsize(Data,0)*(flatten ? 4 : 1) // Usually 32, or 128 if flattened.  
	Variable replications=dimsize(Data,1) // Number of spikes.  
	Variable i,num_PCs=3
	Make /o/n=(replications,num_PCs*num_Electrodes) Lambda
	for(electrode=0;electrode<(flatten ? 1 : num_electrodes);electrode+=1)
		if(flatten)
			Make /o/n=(points,replications) ElectrodeData=Data[mod(p,points/4)][q][floor(mod(p,points)/(points/4))]
		else
			Make /o/n=(points,replications) ElectrodeData=Data[p][q][electrode]
		endif
		PCA /LEIV /SCMT /SRMT ElectrodeData
		Wave M_R // Each column of this matrix is a principal component.  ElectrodeData=M_R*M_C
		Wave M_C // Each row of this matrix is the coefficients for one principal component.  
		Make /o/n=(points,num_PCs) $("PC"+num2str(electrode))=M_R[p][q]
		Make /o/n=(replications,num_PCs) $("Lambda"+num2str(electrode))=M_C[q][p]
		Lambda[][electrode*num_PCs,(electrode+1)*num_PCs-1]=M_C[q-electrode*num_PCs][p] // Lambdas for all tetrodes, with 'num_PCs' columns per tetrode.  
	endfor
	KillWaves /Z M_R,M_C//,ElectrodeData
End

// Extract only spikes from Data that belong to certain clusters.  
Function NlxExtractClusters(Data,type,clusters[,name])
	Wave Data // The wave of spike data.  
	String type // ntt.  
	String clusters // e.g. "8" or "8,12" (8 through 12) or "8,10;14,18" (8 through 10 and 14 through 18).  
	String name // Optional suffix for the extracted data.  Only used when multiple clusters are requested.  
	
	clusters=ListExpand(clusters)
	Wave ClusterAssignments=$(NameOfWave(Data)+"_c")
	Wave Times=$(NameOfWave(Data)+"_t")
	if(ParamIsDefault(name))
		name=clusters[0,7]
	endif
	if(ItemsInList(clusters)==10) // If there is only one cluster to extract, use this faster method.  
		Variable clusterNum=str2num(StringFromList(0,clusters))
		Extract /O Data, $(NameOfWave(Data)+"_"+num2str(clusterNum)) /WAVE=DataCluster, ClusterAssignments[q]==clusterNum
		Extract /O Times, $(NameOfWave(Data)+"_"+num2str(clusterNum)+"_t") /WAVE=DataClusterTimes, ClusterAssignments[p]==clusterNum
	else // We must use the slower method.  
		Make /o/n=(numpnts(ClusterAssignments)) ClusterInclude=WhichListItem(num2str(ClusterAssignments[p]),clusters)>=0
		Extract /O Data, $CleanUpName(NameOfWave(Data)+"_"+name,0) /WAVE=DataCluster, ClusterInclude[q]==1
		Extract /O Times, $CleanUpName(NameOfWave(Data)+"_"+name+"_t",0) /WAVE=DataClusterTimes, ClusterInclude[p]==1
		Extract /O ClusterAssignments, $CleanUpName(NameOfWave(Data)+"_"+name+"_c",0) /WAVE=DataClusterClusters, ClusterInclude[p]==1
		KillWaves /Z ClusterInclude
	endif
	print num2str(numpnts(DataClusterTimes))+" matching spikes."
	strswitch(type)
		case "Tetrode":
			Redimension /n=(32,numpnts(DataCluster)/128,4) DataCluster
			break
	endswitch
End

// Used to cluster Neuralynx spike data using PCA.  Not recommended if something more serious like Klustakwik or BubbleClust, plus manual cutting, is available.  
Function NlxClustering(Lambda,numClasses)
	Wave Lambda
	Variable numClasses
	MatrixOp /O LambdaT=Lambda^t
	KMeans /OUT=2 /NCLS=(numClasses) /SEED=(gnoise(1)) /INIT=1 LambdaT
	Wave W_KMMembers
	//ModifyGraph zColor={W_KMMembers,*,*,Rainbow,0}
	ColorTab2Wave Rainbow; Wave M_Colors
	Variable numColorWavePoints=dimsize(M_Colors,0), numSpikes=dimsize(Lambda,0)
	Make /o/n=(numSpikes,4) PCAColors
	PCAColors=(1/65535)*M_colors[W_KMMembers[p]*numColorWavePoints/numClasses][q]
	PCAColors[][3]=1 // Alpha setting.  
	//Execute /Q "ModifyGizmo /Z ModifyObject=scatter0 property={colorWave,PCAColors}"
	KillWaves /Z LambdaT
End

// Creates a color wave to use for coloring Neuralynx spike data plotted in a Gizmo or scatter plot to review clustering.  
Function NlxColorByCluster(Data)
	Wave Data
	
	Wave Clusters=$(NameOfWave(Data)+"_c")
	ColorTab2Wave Rainbow; Wave M_Colors
	Variable numColorWavePoints=dimsize(M_Colors,0), numSpikes=numpnts(Clusters)
	Make /o/n=(numSpikes,4) ClusterColors
	Make /o/n=100 TempHist
	Histogram /B=2 Clusters, TempHist
	Extract /O/INDX TempHist,TempIndex,TempHist>0
	Variable numClusters=numpnts(TempIndex)
	ClusterColors=(1/65535)*M_colors[FindValue2(TempIndex,Clusters[p])*numColorWavePoints/numClusters][q]
	ClusterColors[][3]=1 // Alpha setting.  
	KillWaves /Z TempHist,TempIndex
End

// ---------------------------------- Simple spike viewer for single electrode data. -----------------------------------------------
Function DisplayAllSpikes(Spikes)
	Wave Spikes
	
	Display /K=1
	Variable i
	for(i=0;i<min(1000,dimsize(Spikes,1));i+=1)
		AppendToGraph /c=(0,0,0) Spikes[][i]
	endfor
	Make /o/n=(i) SpikeTraceMap=p
	ControlBar /T 30
	SetVariable SpikeNum value=_NUM:0, proc=DisplayAllSpikesSetVariables, userData(previousValue)="0"
	Button Reject, proc=DisplayAllSpikesButtons, title="Reject"
	Make /o/n=(i) BadSpikes=0
	SetWindow kwTopWin hook(clickHook)=DisplayAllSpikesHook, userData(numSpikes)=num2str(i)
End

Function DisplayAllSpikesSetVariables(info)
	Struct WMSetVariableAction &info
	if(info.eventCode<0)
		return 0
	endif
	strswitch(info.ctrlName)
		case "SpikeNum":
			Variable spikeNum=info.dval
			String traces=TraceNameList("",";",1)
			String traceStem=StringFromList(0,traces)
			Variable numTraces=ItemsInList(traces)
			String spikeTrace=traceStem+"#"+num2str(spikeNum)
			String previousSpikeTrace=traceStem+"#"+GetUserData("","SpikeNum","previousValue")
			String lastTrace=traceStem+"#"+num2str(numTraces-1)
			ModifyGraph rgb($lastTrace)=(0,0,0)
			if(!StringMatch(lastTrace,spikeTrace))
				ReorderTraces $lastTrace, {$lastTrace,$spikeTrace}
			endif
			ModifyGraph rgb($lastTrace)=(65535,0,0)
			SetVariable SpikeNum userData(previousValue)=num2str(spikeNum)
			break
	endswitch
End

Function DisplayAllSpikesButtons(info)
	Struct WMButtonAction &info
	if(info.eventCode!=2)
		return 0
	endif
	String traces=TraceNameList("",";",1)
	String stem=StringFromList(0,traces)
	String lastTrace=StringFromList(ItemsInList(traces)-1,traces)
	strswitch(info.ctrlName)
		case "Reject":
			ControlInfo SpikeNum; Variable currSpikeNum=V_Value
			ModifyGraph hideTrace($lastTrace)=1
			Wave BadSpikes
			Variable spikeNum=WaveColumnFromTraceName(lastTrace)
			BadSpikes[spikeNum]=1
			break
	endswitch
End

// Removes bad spikes added using the "reject" button.  
Function RemoveBadSpikes(Data,BadSpikes)
	Wave Data,BadSpikes
	Variable i
	for(i=numpnts(BadSpikes)-1;i>=0;i-=1)
		if(BadSpikes[i]==1)
			DeletePoints /M=1 i,1,Data
		endif
	endfor
End

// Hook function for Display all spikes.  
Function DisplayAllSpikesHook(info)
	STRUCT WMWinHookStruct &info
	if(info.eventCode==5) // Mouse up.  
		String traceClicked=StringByKey("TRACE",TraceFromPixel(info.mouseLoc.h,info.mouseLoc.v,""))
		Variable instance
		sscanf traceClicked,"%*[a-zA-Z0-9]#%d",instance
		STRUCT WMSetVariableAction info2
		info2.ctrlName="SpikeNum"
		info2.dval=instance
		SetVariable SpikeNum value=_NUM:instance
		DisplayAllSpikesSetVariables(info2)
	endif
End

// ------------------------------

// ---------------------- Begin non-Nlx functions (to put .ibw data into the analysis format used for Nlx files) ------------------------------------ //

// Exports all of the igor data waves as one or more concatenated data waves.  
Function ExportChannelData(channel [,stem,folder,Timestamps,max_piece_size,sayv,kill,hi])
	String channel,stem,folder
	Wave /Z Timestamps
	Variable max_piece_size // In x-scaling units (usually seconds).  Set this if you want to analyze smaller chunks of continuous data.  
	Variable sayv // Save to disk (as .ibw).  
	Variable kill // Kill concatentations when finished (not original data).  
	Variable hi // High-pass the exported data at this cutoff frequency.  
	
	max_piece_size=ParamIsDefault(max_piece_size) ? inf : max_piece_size
	sayv=ParamIsDefault(sayv) ? 1 : sayv
	kill=ParamIsDefault(kill) ? 1 : kill
	if(ParamIsDefault(stem))
		stem="sweep"
	endif
	if(ParamIsDefault(folder))
		folder="root"
	endif
	if(ParamIsDefault(Timestamps))
		Wave Timestamps=root:SweepT
	endif
	
	String curr_folder=GetDataFolder(1)
	SetDataFolder $(folder+":"+channel)
	String data_wave_names=WaveList(stem+"*",";","")
	Variable i,data_num,cumul_data_wave_duration=0, num_data_waves=ItemsInList(data_wave_names),epoch=0,piece=0,last_timestamp=Inf
	String data_wave_list=""
	NewPath /O/C/Q SaveLocation,SpecialDirPath("Desktop",0,0,0)+IgorInfo(1)
	
	for(i=0;i<num_data_waves;i+=1)
		String data_wave_name=StringFromList(i,data_wave_names)
		Wave DataWave=$data_wave_name
		sscanf data_wave_name,stem+"%d",data_num
		Variable timestamp=Timestamps[data_num-1] // e.g. sweep1 has its start time in Timestamps[0].  
		Variable data_wave_duration=dimdelta(DataWave,0)*dimsize(DataWave,0)
		if(data_wave_duration>max_piece_size)
			print data_wave_name+" was larger than the maximum piece size."
			return -1
		endif
		if(last_timestamp+data_wave_duration<timestamp) // Non-contiguous recording.  
			ConcatPiece(channel,data_wave_list,epoch,piece,cumul_data_wave_duration-data_wave_duration,last_timestamp,sayv,kill,hi)
			Note $(channel+num2str(epoch)+"_"+num2str(piece))
			cumul_data_wave_duration=data_wave_duration
			epoch+=1
			piece=0
			data_wave_list=data_wave_name+";" // Start a new list for concatenation.  
			last_timestamp=timestamp
		else // This beginning of this data wave is temporally contiguous with the end of the last.  
			if(cumul_data_wave_duration+data_wave_duration>max_piece_size) // The next concatenation would be too big.  
				ConcatPiece(channel,data_wave_list,epoch,piece,cumul_data_wave_duration,timestamp,sayv,kill,hi) // Make a data wave piece.  
				cumul_data_wave_duration=0
				data_wave_list=""
				piece+=1
				i-=1 // Process this data wave again in the next iteration of the loop.  
			else
				cumul_data_wave_duration+=data_wave_duration
				last_timestamp=timestamp
				data_wave_list+=data_wave_name+";" // Add this data wave to the list for concatenation.  
			endif
		endif
	endfor
	if(cumul_data_wave_duration>0)
		ConcatPiece(channel,data_wave_list,epoch,piece,cumul_data_wave_duration-data_wave_duration,timestamp,sayv,kill,hi)
	endif
	
	SetDataFolder $curr_folder
End

// Concatenates pieces of data.  
Function ConcatPiece(channel,wave_list,epoch,piece,cumul_duration,timestamp,sayv,kill,hi)
	String channel,wave_list
	Variable epoch,piece,cumul_duration,timestamp,sayv,kill,hi
	String concat_name=channel+num2str(epoch)+"_"+num2str(piece)
	Concatenate /O/NP wave_list, $concat_name; Wave Concat=$concat_name
	if(hi>0)
		Variable offset=Concat[0]
		Concat-=offset // Must do this to prevent filtering from producing a ringing transient.  
		FilterIIR /HI=(hi) Concat
		Concat+=offset
		Note Concat, "FilterIIR:"+num2str(hi/dimdelta(Concat,0))
	endif
	Note Concat, "Timestamp:"+num2str(timestamp-cumul_duration)
	if(sayv)
		Save /P=SaveLocation Concat as concat_name+".ibw"
	endif
	if(kill)
		KillWaves /Z Concat
	endif
End

// Extract only the parts of the data that contains the spikes.  Used on continuous data to extract peri-spike segments.   
Function ExtractChannelSpikes(channel,threshold[,polarity,epoch,winLeft,winRight])
	String channel
	Variable threshold,polarity,epoch,winLeft,winRight // polarity is 0 for up, 1 for down.  
	
	winLeft=ParamIsDefault(winLeft) ? 0.0003 : winLeft
	winRight=ParamIsDefault(winRight) ? 0.0012 : winRight
	
	String epochStr=SelectString(ParamIsDefault(epoch),num2str(epoch)+"*","*")
	String dataPieceNames=WaveList(channel+epochStr,";","")
	Variable i
	for(i=0;i<ItemsInList(dataPieceNames);i+=1)
		String dataPieceName=StringFromList(i,dataPieceNames)
		Wave DataPiece=$dataPieceName
		ExtractPieceSpikes(DataPiece,threshold,polarity,winLeft,winRight)
	endfor
End

// Auxiliary function for ExtractChannelSpikes, used on one piece of the data recorded on a channel.  
Function ExtractPieceSpikes(DataPiece,threshold,polarity,winLeft,winRight)
	Wave DataPiece
	Variable threshold,polarity,winLeft,winRight
	Variable epoch
	sscanf NameOfWave(DataPiece),"%*[A-Za-z]%d", epoch
	String spikesName,spikeTimesName
	sprintf spikesName "Spikes%d", epoch
	sprintf spikeTimesName "Spikes%d_t", epoch
	
	Make /o/n=0 $spikeTimesName
	FindLevels /EDGE=(polarity+1) /D=$spikeTimesName DataPiece, threshold
	Wave SpikeTimes=$spikeTimesName
	
	Variable i
	Variable rows=(winLeft+winRight)/dimdelta(DataPiece,0)
	Variable cols=numpnts(SpikeTimes)
	
	Make /o/n=(rows,cols) $spikesName; Wave Spikes=$spikesName
	for(i=0;i<cols;i+=1)
		Variable spikeTime=SpikeTimes[i]
		Variable startPoint=x2pnt(DataPiece,spikeTime-winLeft)
		Spikes[][i]=DataPiece[startPoint+p]
	endfor
End

// Delete spikes that are likely shadows of other spikes (ISI < 750 microseconds).  
Function DeleteShadowSpikes(SpikeWave,TimeWave)
	Wave SpikeWave,TimeWave
	Variable i=1
	Do
		if(TimeWave[i+1]-TimeWave[i]<750)
			DeletePoints i,1,TimeWave
			DeletePoints /M=1 i,1,SpikeWave
		else
			i+=1
		endif
	While(i<dimsize(SpikeWave,1))
End

// Makes a spike-time histogram from loaded Neuralynx spike data.  
Function /S MakeNlxSTH(Data,binWidth[,tStart,tStop])
	Wave Data
	Variable binWidth // Width of bins in seconds.  
	Variable tStart,tStop // Optional start and end times to search for spikes, in seconds.  
	
	String name=NameOfWave(Data)
	Wave Times=$(name+"_t")
	Wave Clusters=$(name+"_c")
	
	tStart=ParamIsDefault(tStart) ? wavemin(Times) - 0.001 : max(tStart,0) // Start time in seconds.  
	tStop=ParamIsDefault(tStop) ? wavemax(Times) + 0.001 : min(tStop,wavemax(Times)+100) // Stop time in seconds.  
	Variable i,maxCluster=wavemax(Clusters)
	
	// Count number of distinct clusters in the cluster wave.  
	Make /o/n=100 TempHist // Maximum of 100 clusters.  
	Histogram /B=2 Clusters,TempHist
	TempHist=TempHist>0 && p!=0 // 1's for all clusters with spikes except for the unassigned cluster.  
	Variable numClusters=sum(TempHist)
	Variable numBins=ceil(tstop-tstart)/(binWidth)
	
	Make /o/n=(numBins,numClusters) $(NameOfWave(Data)+"_sth") /WAVE=NlxSTH=0
	for(i=1;i<=maxCluster;i+=1) // Start from i=1 to exclude unclustered spikes.  
		Make /o/n=(numBins) TempHist=0
		SetScale x,tStart,tStart+numBins*binWidth,TempHist,NlxSTH
		Extract /O Times,$(name+"_t_"+num2str(i)) /WAVE=OneClusterTimes,Clusters[p]==i
		if(numpnts(OneClusterTimes))
			Histogram /B=2 OneClusterTimes, TempHist
			NlxSTH[][i]=TempHist[p]
		endif
	endfor
	SetScale /P x,tStart,binWidth,NlxSTH
	NlxSTH/=binWidth
	KillWaves /Z TempHist
	return GetWavesDataFolder(NlxSTH,2)
End

// Makes a peri-stimulus spike-time histogram from a regular spike-time histogram.  TriggerTimes contains a list of the stimuli or events to use as the 0 time bins for the histogram.  
Function /S MakeNlxPSTH(NlxSTH,TriggerTimes,preTrig,postTrig)
	Wave NlxSTH,TriggerTimes
	Variable preTrig,postTrig // e.g. -3 and 10.  
	
	Variable i
	Variable numBins=ceil((postTrig+preTrig)/dimdelta(NlxSTH,0))
	String baseName=RemoveEnding(NameOfWave(NlxSTH),"_sth")
	Make /o/n=(numBins,dimsize(NlxSTH,1)) $(baseName+"_psth") /WAVE=NlxPSTH=0, $(baseName+"_psth_sem") /WAVE=NlxPSTH_SEM=0
	SetScale x,-preTrig,postTrig,NlxPSTH,NlxPSTH_SEM
	for(i=0;i<numpnts(TriggerTimes);i+=1)
		NlxPSTH+=NlxSTH(x+TriggerTimes[i])[q]
		NlxPSTH_SEM+=(NlxSTH(x+TriggerTimes[i])[q])^2
	endfor
	NlxPSTH/=i
	NlxPSTH_SEM=sqrt(NlxPSTH_SEM/i-NlxPSTH^2)/sqrt(i)
	Duplicate /o NlxPSTH, $(baseName+"_psth_low") /WAVE=NlxPSTH_low, $(baseName+"_psth_high") /WAVE=NlxPSTH_high
	NlxPSTH_low-=NlxPSTH_SEM
	NlxPSTH_high+=NlxPSTH_SEM
	return GetWavesDataFolder(NlxPSTH,2)
End

Function DisplayNlxSTH(NlxSTH)
	Wave NlxSTH
	
	Variable binWidth=dimdelta(NlxSTH,0)
	String name=NameOfWave(NlxSTH)
	NlxSTH=log(NlxSTH+(1/binWidth)) // Shifted log scale.  
	NewImage /K=1 NlxSTH
	Make /o/n=4 $(name+"_tks") /WAVE=Ticks_={0,0.1,1,10}; 
	Ticks_=log(Ticks_+1/binWidth)
	Make /o/T/n=4 $(name+"_tksl") /WAVE=TickLabels=num2str((10^Ticks_)-1/binWidth)
	ModifyImage $name ctab={log(1/binWidth),log(10/binWidth+1/binWidth),ColdWarm,0}
	ColorScale userTicks={Ticks_,TickLabels}
	MoveWindow 100,100,800,500
End

Function DisplayNlxPSTH(NlxPSTH,NlxPSTH_low,NlxPSTH_high)
	Wave NlxPSTH,NlxPSTH_low,NlxPSTH_high
	
	Display /K=1 /W=(0,0,500,800)
	Variable i,unit,numUnits=dimsize(NlxPSTH,1)
	Make /o/n=0 TempActiveUnits
	
	for(i=0;i<numUnits;i+=1)
		ImageStats /Q/G={0,dimsize(NlxPSTH,0)-1,i,i} NlxPSTH
		if(V_max>0)
			TempActiveUnits[numpnts(TempActiveUnits)]={i}
		endif
	endfor
	numUnits=numpnts(TempActiveUnits)
	for(i=0;i<numUnits;i+=1)
		unit=TempActiveUnits[i]
		String axisName="axis_"+num2str(unit)
		AppendToGraph /L=$axisName NlxPSTH_low[][unit]
		String last=TopTrace()
		ModifyGraph mode($last)=7, toMode($last)=1, hbFill($last)=4
		AppendToGraph /L=$axisName NlxPSTH[][unit]
		last=TopTrace()
		ModifyGraph mode($last)=7, toMode($last)=1, hbFill($last)=4, lsize($last)=2
		AppendToGraph /L=$axisName NlxPSTH_high[][unit]
		ModifyGraph /Z axisEnab($axisName)={i/numUnits+min(0.01,1/numUnits),(i+1)/numUnits-min(0.01,1/numUnits)}, freePos($axisName)={0.05,kwFraction}
		ImageStats /Q/G={0,dimsize(NlxPSTH,0)-1,unit,unit} NlxPSTH
		SetAxis /Z $axisName, 0, V_max*1.5
	endfor
	ModifyGraph axisEnab(bottom)={0.05,1}, zero(bottom)=2
	KillWaves /Z TempActiveUnits
End

// Uses a continuously recorded Neuralynx wave to search for e.g. stimulus artifacts that could indicate stimulus times in the absence of a proper time series.  
Function NlxStimEpochs(ContinuousWave,threshold)
	Wave ContinuousWave // A continuous recording that will have stimulus artifacts that can be identified by a level crossing.  
	Variable threshold // The level crossing.  
	
	Variable expectedInterval=20
	Variable tolerance=0.5
	Variable minInterval=expectedInterval-tolerance
	
	FindLevels /Q /M=(minInterval) /EDGE=1 ContinuousWave,threshold
	Wave W_FindLevels
	Variable i,epochNum=0
	for(i=0;i<numpnts(W_FindLevels);i+=1)
		if(abs(W_FindLevels[i]-W_FindLevels[i-1]-expectedInterval)>tolerance)
			Make /o/n=0 $("Epoch"+num2str(epochNum)) /WAVE=Epoch
			epochNum+=1
		endif
		Epoch[numpnts(Epoch)]={W_FindLevels[i]}
	endfor
	KillWaves /Z W_FindLevels
End

// ---------------- Begin Neuralynx Panel Functions ---------------------

Menu "Analysis"
	"Neuralynx",/Q,NeuralynxPanel()
End

Window NeuralynxPanel() : Panel
	PauseUpdate; Silent 1		// building window...
	DoWindow /K NeuralynxPanel
	NewPanel /K=1 /W=(577,90,1158,199)
	
	GroupBox SaveLoad,pos={1,0},size={485,47}
	SetVariable dirName,pos={3,2},size={288,16},proc=NeuralynxPanelSetVariables,title="Directory"
	PathInfo NeuralynxPath
	SetVariable dirName,value= _STR:SelectString(V_flag,"",S_path)
	SetVariable fileName,pos={93,25},size={197,16},title="File",value= _STR:""
	PopupMenu Type,pos={3,23},size={83,21},title="Type"
	PopupMenu Type,mode=1,popvalue="ncs",value= #"\"ncs;ntt\""
	Button ChooseDir,pos={292,1},size={19,20},proc=NeuralynxPanelButtons,title="..."
	Button ChooseFile,pos={292,24},size={19,19},proc=NeuralynxPanelButtons,title="..."
	Button Load,pos={321,1},size={50,20},proc=NeuralynxPanelButtons,title="Load"
	SetVariable Downsample,value=_NUM:0,size={100,20},limits={0,Inf,1},title="Downsample"
	Button Save,pos={321,24},size={50,20},proc=NeuralynxPanelButtons,title="Save"
	
	GroupBox SaveLoad1,pos={1,51},size={574,53}
	PopupMenu Data,pos={3,53},size={83,21},proc=NeuralynxPanelPopupMenus,title="Data"
	PopupMenu Data,mode=1,popvalue="ntt2",value= #"NlxAvailableWaves(\"ntt\")"
	Button View,proc=NeuralynxPanelButtons, title="View"
	SetVariable tStart,size={71,16},title="Start",value= _NUM:0
	SetVariable tStop,size={75,16},title="Stop",value= _NUM:inf
	SetVariable binWidth,size={75,16},title="Width",value= _NUM:1
	CheckBox DisplaySTH,size={45,14},title="Show",value= 0
	Button MakeSTH,size={50,20},proc=NeuralynxPanelButtons,title="STH"
	
	PopupMenu STH,pos={3,79},size={97,21},proc=NeuralynxPanelPopupMenus,value=#"WaveList(\"*_sth\",\";\",\"\")",title="STH"
	PopupMenu TriggerTimes,size={87,21},proc=NeuralynxPanelPopupMenus,title="Trig"
	PopupMenu TriggerTimes,mode=1,popvalue="ntt2_t",value= #"WaveList(\"*\",\";\",\"DIMS:1\")"
	SetVariable tPre,size={65,16},title="Pre",value= _NUM:-1
	SetVariable tPost,size={75,16},title="Post",value= _NUM:10
	CheckBox DisplayPSTH,size={45,14},title="Show",value= 0
	Button MakePSTH,size={50,20},proc=NeuralynxPanelButtons,title="PSTH"
EndMacro

Function NeuralynxPanelSetVariables(info)
	Struct WMSetVariableAction &info
	if(info.eventCode != 1 && info.eventCode != 2)
		return -1
	endif
	strswitch(info.ctrlName)
		
	endswitch
End

Function NeuralynxPanelButtons(ctrlName)
	String ctrlName
	strswitch(ctrlName)
		case "ChooseDir":
			NewPath /M="Choose a directory of Neuralynx files..."/O/Q/Z NeuralynxPath
			if(!V_flag)
				PathInfo NeuralynxPath
				SetVariable dirName, value=_STR:S_path
			else
				return -2
			endif
			break
		case "ChooseFile":
			Variable refNum
			//PathInfo NeuralynxPath
			ControlInfo Type; String type=S_Value
			Open /D/R/T=("."+type)/Z=2/P=NeuralynxPath/M="Choose a Neuralynx file..." refNum
			if(strlen(S_fileName))
				String fileName=StringFromList(ItemsInList(S_fileName,":")-1,S_fileName,":")
				SetVariable fileName, value=_STR:RemoveEnding(fileName,"."+type)
			endif
			break
		case "Load":
			ControlInfo Type; type=S_Value
			ControlInfo fileName; fileName=S_Value
			ControlInfo DownSample; Variable downSample=V_Value
			LoadNlxBinaryFile(type,fileName,downSamp=downSample)
			break
		case "View":
			ControlInfo Data; Wave Data=$S_Value
			type=NlxType(Data)
			CreateNlxViewer(Data,type)
			break
		case "MakeSTH":
			ControlInfo tStart; Variable tStart=V_Value
			ControlInfo tStop; Variable tStop=V_Value
			ControlInfo Data; Wave Data=$S_Value
			ControlInfo binWidth; Variable binWidth=V_Value
			String NlxSTHname=MakeNlxSTH(Data,binWidth,tStart=tStart,tStop=tStop)
			ControlInfo DisplaySTH
			if(V_Value)
				DisplayNlxSTH($NlxSTHname)
			endif
			break
		case "MakePSTH":
			ControlInfo STH; NlxSTHname=S_Value
			ControlInfo TriggerTimes; String TriggerTimesname=S_Value
			ControlInfo tPre; Variable tPre=V_Value
			ControlInfo tPost; Variable tPost=V_Value
			String NlxPSTHname=MakeNlxPSTH($NlxSTHname,$TriggerTimesname,tPre,tPost)
			ControlInfo DisplayPSTH
			if(V_Value)
				DisplayNlxPSTH($NlxPSTHname,$(NlxPSTHname+"_low"),$(NlxPSTHname+"_high"))
			endif
			break
	endswitch
End

Function NeuralynxPanelPopupMenus(info)
	Struct WMPopupAction &info
End

// ----------- Begin static helper functions ------------------

// Returns index of the the numeric value 'value' or -1 if not found.  
static Function FindValue2(Wav,value)
	Wave Wav
	Variable value
	FindValue /V=(value) Wav
	return V_Value
End

// Returns the name of the top-most trace on a graph.  
static Function /S TopTrace([win,xaxis,yaxis,visible,number])
	String win,xaxis,yaxis
	Variable visible,number
	if(ParamIsDefault(win))
		win=WinName(0,1)
	endif
	String traces=TraceNameList(win,";",3)
	Variable i
	for(i=ItemsInList(traces)-1;i>=0;i-=1)
		String trace=StringFromList(i,traces)
		Variable match=1
		String trace_info=TraceInfo(win,trace,0)
		if(!ParamIsDefault(xaxis))
			String traceXAxis=StringByKey("XAXIS",trace_info)
			match*=StringMatch(xaxis,traceXAxis)
		endif
		if(!ParamIsDefault(yaxis))
			String traceYAxis=StringByKey("YAXIS",trace_info)
			match*=StringMatch(yaxis,traceYAxis)
		endif
		if(!ParamIsDefault(visible))
			Variable hidden=str2num(StringByKey("hideTrace(x)",trace_info,"="))
			match*=(visible!=hidden)
		endif
		if(match)
			if(number<=0)
				return trace
			else
				number-=1
			endif
		endif
	endfor
	return ""
End

// Expands a list of the form "3,6;12,14" to "3;4;5;6;12;13;14"
static Function /S ListExpand(list)
	String list
	String new_list=""
	Variable i,j
	String sub_list,first,last,item
	for(i=0;i<ItemsInList(list);i+=1)
		sub_list=StringFromList(i,list)
		first=StringFromList(0,sub_list,",")
		last=StringFromList(1,sub_list,",")
		if(!cmpstr("",last))
			new_list=AddListItem(first,new_list)
		else
			for(j=str2num(first);j<=str2num(last);j+=1)
				item=num2str(j)
				new_list=AddListItem(item,new_list)
			endfor
		endif
	endfor
	return ReverseListOrder(new_list)
End

// Reverses the order of the items in a string list.  
static Function /S ReverseListOrder(list)
	String list
	String new_list=""
	Variable i
	String item
	for(i=0;i<ItemsInList(list);i+=1)
		item=StringFromList(i,list)
		new_list=AddListItem(item,new_list)
	endfor
	return new_list
End

// Returns the column of a wave used to plot a given trace.  
static Function WaveColumnFromTraceName(traceName[,win])
	String traceName,win
	if(ParamIsDefault(win))
		win=WinName(0,1)
	endif
	String info=TraceInfo(win,traceName,0)
	String yRange=StringByKey("YRANGE",info)
	Variable column
	sscanf yRange,"[%*[*0-9,]][%d]",column
	return column
End